// takaha.c
//
// Copyright © 2023 Electric Joseph
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the “Software”), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#include <windows.h>

const wchar_t takaha[] = L"takaha";
wchar_t **files;
int n_files;
wchar_t **texts, ***pages;
int n_texts;
int current_page;
int wcx, wcy; // window size
int opt_x;

wchar_t **
read_files(wchar_t **files, int n_files, int *n_texts)
{
   if (!files || !*files) return NULL;

   wchar_t **r = NULL;
   wchar_t *l  = NULL;
   int nl = 0;
   int nr = 0;
   wchar_t c;
   int q = 0;

   for (int i = 0; i < n_files; i++) {
      wchar_t *file = files[i];
      if (file[0] == L'-') {
         for (int i = 0; file[i]; i++)
            if (file[i] == L'x') opt_x = 1;
         continue; }

      HANDLE f;
      f = CreateFile(file, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
      if (f == INVALID_HANDLE_VALUE) continue;

      for (;;) {
         char t;
         unsigned long un;
         ReadFile(f, &t, sizeof t, &un, NULL);
         if (!un) t = '\n';

         // UTF-8
         if (t > 0)
            c = t; // ascii
         else if ((t & 0xc0) == 0x80 && q > 0) {
            c = (c << 6) | (t & 0x3f); if (--q) continue; }
         else if ((t & 0xe0) == 0xc0) {
            c = t & 0x1f, q = 1; continue; }
         else if ((t & 0xf0) == 0xe0) {
            c = t & 0x0f, q = 2; continue; }
         else
            c = '?'; // replace head

         nl++;
         l = realloc(l, nl * sizeof(*l)); if (!l) return NULL;
         l[nl - 1] = t == '\n' ?  L'\0' : c;

         if (t == '\n') {
            nr++;
            r = realloc(r, nr * sizeof(*r)); if (!r) return NULL;
            r[nr - 1] = l;
            l = NULL;
            nl = 0;
         }
         if (!un) break;
      }
      CloseHandle(f);
   }

   *n_texts = nr;
   return r;
}

wchar_t ***
texts_to_pages(wchar_t **texts, const int n)
{
   wchar_t ***p = NULL;
   int c = 0, // page count
       s = 0; // state
   for (int i = 0; i < n; i++) {
      int t = !!texts[i] && texts[i][0];
      if (!s && t) { // new page
         p = realloc(p, ++c * sizeof(*p));
         p[c - 1] = &texts[i]; }
      s = t; }
   // printf("%s: %d pages\n", __func__, c); // DEBUG
   p = realloc(p, ++c * sizeof(*p));
   p[c - 1] = NULL;
   return p;
}

SIZE
calc_text_size(HDC hdc, int fontsize, const wchar_t *text, int flag)
{
   LOGFONT lf = { .lfHeight = fontsize, };
   HFONT fnt = CreateFontIndirect(&lf), old_fnt = NULL;
   if (fnt) old_fnt = SelectObject(hdc, fnt);
   const int l = wcslen(text);
   SIZE size;
   GetTextExtentPoint32(hdc, text, l, &size);
   if (flag) {
      int x = (wcx - size.cx) / 2;
      int y = flag;
      TextOut(hdc, x, y, text, l); }
   if (old_fnt) SelectObject(hdc, old_fnt);
   if (fnt) DeleteObject(fnt);
   return size;
}

LRESULT CALLBACK
wp_paint(HWND hwnd, UINT mesg, WPARAM wp, LPARAM lp)
{
   PAINTSTRUCT ps;
   HDC hdc = BeginPaint(hwnd, &ps);
   wchar_t **t = pages[current_page];
   int n = &texts[n_texts] - t;

   double x_factor = -1.0;
   if (opt_x)
      for (int i = 0; i < n; i++) {
         if (!t[i] || !t[i][0]) break;
         SIZE size = calc_text_size(hdc, 240, t[i], 0);
         double factor = (double)wcx / size.cx;
         if (x_factor < 0 || x_factor > factor)
            x_factor = factor; }

   int y = 8;
   for (int i = 0; i < n; i++) {
      if (!t[i] || !t[i][0]) break;
      SIZE size = calc_text_size(hdc, 240, t[i], 0);
      double factor = (double)wcx / size.cx;
      if (opt_x) factor = x_factor;
      size = calc_text_size(hdc, (int)(factor * 240), t[i], 0);
      y += size.cy; }

   const double y_factor = y > wcy ? (double)wcy / y : 1.0;
   y = y > wcy ? 8 : (wcy - y) / 2 ?: 8;

   for (int i = 0; i < n; i++) {
      if (!t[i] || !t[i][0]) break;
      SIZE size = calc_text_size(hdc, 240, t[i], 0);
      double factor = (double)wcx / size.cx;
      if (opt_x) factor = x_factor;
      size = calc_text_size(hdc, (int)(y_factor * factor * 240), t[i], y);
      y += size.cy; }

   EndPaint(hwnd, &ps);
}

LRESULT CALLBACK
wproc(HWND hwnd, UINT mesg, WPARAM wp, LPARAM lp)
{
   switch (mesg) {
   case WM_LBUTTONDOWN:
      current_page++;
      if (!pages[current_page])
         current_page = 0;
      InvalidateRect(hwnd, NULL, TRUE);
      return 0;
   case WM_RBUTTONDOWN:
      if (current_page > 0)
         current_page--;
      InvalidateRect(hwnd, NULL, TRUE);
      return 0;
   case WM_DESTROY:
      PostQuitMessage(0);
      return 0;
   case WM_SIZE:
      wcx = LOWORD(lp);
      wcy = HIWORD(lp);
      InvalidateRect(hwnd, NULL, TRUE);
      return 0;
   case WM_PAINT:
      return wp_paint(hwnd, mesg, wp, lp);
   default:
      return DefWindowProc(hwnd, mesg, wp, lp); }
}
#include <stdio.h>

int WINAPI
wWinMain(HINSTANCE hinst,
      HINSTANCE hprev,
      wchar_t *cmdline,
      int cmdshow)
{
   if (!*cmdline) return 1;
   files = CommandLineToArgvW(cmdline, &n_files);
   if (!files) return 1;
   texts = read_files(files, n_files, &n_texts);
   if (!texts) return 1;
   pages = texts_to_pages(texts, n_texts);
   if (!pages) return 1;

   WNDCLASS wc = {
      .lpfnWndProc = wproc,
      .hInstance = hinst,
      .hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH),
      .lpszClassName = takaha, };
   if (!RegisterClass(&wc)) return 1;

   HWND wnd = CreateWindow(takaha, takaha,
         WS_OVERLAPPEDWINDOW,
         CW_USEDEFAULT,
         CW_USEDEFAULT,
         CW_USEDEFAULT,
         CW_USEDEFAULT,
         0, 0, hinst, NULL);
   if (!wnd) return 1;
   ShowWindow(wnd, cmdshow);

   MSG mesg;
   while (GetMessage(&mesg, 0, 0, 0))
      DispatchMessage(&mesg);

   return mesg.wParam;
}
// vim:sw=3:
